/*
  DUST: Simulator for circuits containing a SX microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//---------------------------------------------------------------------------
#include "basecomponents.h"

/*******************************************************************************
Classe NOT
*******************************************************************************/
void not::Update(int force)
{
  O.SetState(!(A.GetState()));
  updateneeded = 0;
  updatetime = +INFINITY;
}

not::not(componenttype *type, int name, attribute *attribs, symbol *parent)
  : component(type, name, attribs, parent)
  , A(symbtable.toid("A"), this, attribs, this)
  , O(symbtable.toid("O"), this, attribs, this)
{
  updatetime = +INFINITY;
  parseattribs(attribs);
}


/*******************************************************************************
Classe AND 
*******************************************************************************/
void and::Update(int force)
{
  O.SetState((A.GetState() && B.GetState()));
  updateneeded = 0;
  updatetime = +INFINITY;
}

and::and(componenttype *type, int name, attribute *attribs, symbol *parent)
  : not(type, name, attribs, parent),
  B(symbtable.toid("B"), this, attribs, this)
{
}

/*******************************************************************************
Classe OR
*******************************************************************************/
void or::Update(int force)
{
  O.SetState((A.GetState() || B.GetState()));
  updateneeded = 0;
  updatetime = +INFINITY;
}

/*******************************************************************************
Classe XOR
*******************************************************************************/
void xor::Update(int force)
{
  O.SetState((A.GetState()^B.GetState()));
  updateneeded = 0;
  updatetime = +INFINITY;
}

/*******************************************************************************
Classe NAND
*******************************************************************************/
void nand::Update(int force)
{
  O.SetState(!(A.GetState() && B.GetState()));
  updateneeded = 0;
  updatetime = +INFINITY;
}

/*******************************************************************************
Classe NOR
*******************************************************************************/
void nor::Update(int force)
{
  O.SetState(!(A.GetState() || B.GetState()));
  updateneeded = 0;
  updatetime = +INFINITY;
}

/*******************************************************************************
Classe XNOR
*******************************************************************************/
void xnor::Update(int force)
{
  O.SetState(!(A.GetState()^B.GetState()));
  updateneeded = 0;
  updatetime = +INFINITY;
}

/*******************************************************************************
powerdc
*******************************************************************************/

powerdc::powerdc(componenttype *type, int name, attribute *attribs, symbol *parent) : component(type, name, attribs, parent), O(symbtable.toid("VCC"), this, attribs, this)
{
  updateneeded = 0;
  updatetime = +INFINITY;
  O.SetState(1);
  parseattribs(attribs);
}

void powerdc::parseattribs(attribute *attribs)
{
  int state = O.GetState();;
  attribs->getintattribute(symbtable.toid("STATE"), &state);
  O.SetState(state);
}

void powerdc::Update(int force)
{
  //  DEBUG(cerr << "Update powerdc component : " << symbtable.getsymbol(name) << "\n";)

  updateneeded = 0;
  updatetime = +INFINITY;
}

/*******************************************************************************
clock
*******************************************************************************/

Clock::Clock(componenttype *type, int name, attribute *attribs, symbol *parent) : component(type, name, attribs, parent), O(symbtable.toid("CK"), this, attribs, this)
{
  updatetime = Project.GetTime();
  updateneeded = 0;
  parseattribs(attribs);
  _Reset();
}

void Clock::parseattribs(attribute *attribs)
{
  real f = 1e6;
  int freqpos = attribs->getrealattribute("f", &f);
  if (f <= 0) 
  {
    error(TYPE_ERROR, -1, NULL, "Negative frequency requested for clock %s. 1MHz used.");
    T = 1e-6;
  }
  else
  T = 1 / f;

  if (attribs->getrealattribute("T", &T, freqpos));

  if (T <= 0) 
  {
    error(TYPE_ERROR, -1, NULL, "Negative period requested for clock %s. 1MHz used.");
    T = 1e-6;
  }
}

void Clock::_Reset()
{
  O.SetState(0);
  updatetime = 0;
}

void Clock::Update(int force)
{
  if (Project.GetTime() >= updatetime)
  {
    O.SetState(!O.GetState());
    updatetime += T / 2;
  }

  updateneeded = 0;

  //  DEBUG(cerr << "Update clock component : " << symbtable.getsymbol(name) << " UpdateTime is " << updatetime << "\n";)
}

symbol *Clock::_getlocalsymbol(list<int> *name)
{
  if (*name->elem == symbtable.toid("T"))
    return (new realptr(symbtable.toid("T"), getattributes(), this, &T))->getlocalsymbol(name->next);

  return (symbol *) NULL;
}

list<symbol> *Clock::getapparentsymbols()
{
  int name;
  list<int> l(&name);

  name = symbtable.toid("T");
  getlocalsymbol(&l);

  return children;
}

/*******************************************************************************
flipflop
*******************************************************************************/

void flipflop::Update(int force)
{
  if (S.GetState()) state = 1;
  if (R.GetState()) state = 0;
  if (state)
  {
    Q.SetState(1);
    NQ.SetState(0);
    S.care = 0;
    R.care = 1;
  }
  else
  {
    Q.SetState(0);
    NQ.SetState(1);
    S.care = 1;
    R.care = 0;
  }
  updateneeded = 0;
  updatetime = +INFINITY;
}

flipflop::flipflop(componenttype *type, int name, attribute *attribs, symbol *parent) : component(type, name, attribs, parent), S(symbtable.toid("S"), this, attribs, this), R(symbtable.toid("R"), this, attribs, this), Q(symbtable.toid("Q"), this, attribs, this), NQ(symbtable.toid("NQ"), this, attribs, this), svar(symbtable.toid("STATE"), attribs, this, &state, 0, 0)
{
  state = 0;
  updatetime = +INFINITY;  
}

/*******************************************************************************
motor
*******************************************************************************/

motor::motor(componenttype *type, int name, attribute *attribs, symbol *parent) :
component(type, name, attribs, parent),
quad1(symbtable.toid("Q1"), this, attribs, this),
quad2(symbtable.toid("Q2"), this, attribs, this),
pwm(symbtable.toid("PWM"), this, attribs, this),
dir(symbtable.toid("DIR"), this, attribs, this)
{
  _Reset();
}

void motor::_Reset()
{
  x = v = a = extforce = 0;
  A = B = C = 1;
  simulupdatetime = timestep = .0001;  
  quadupdatetime = updatetime = 0;
  parseattribs(getattributes());
}

void motor::Update(int force)
{
  if (Project.GetTime() >= simulupdatetime)
  {
    a = A * pwm.GetVin() * (1 - 2 * dir.GetState()) - B * v + extforce;
    v += a * timestep;
    x += v * timestep;

    simulupdatetime += timestep;
  }
  
  int ix = (int) (C * x);
  quad1.SetState((ix ^ (ix >> 1)) & 1);
  quad2.SetState((ix >> 1) & 1);

  quadupdatetime = Project.GetTime() + C / fabs(v);

  //  cout << quadupdatetime << " " << simulupdatetime << "\n";

  updateneeded = 0;
  updatetime = MIN(simulupdatetime, quadupdatetime);
}

symbol *motor::_getlocalsymbol(list<int> *name)
{
  if (*name->elem == symbtable.toid("Pos"))
    return (new realptr(symbtable.toid("Pos"), getattributes(), this, &x))->getlocalsymbol(name->next);

  if (*name->elem == symbtable.toid("Vel"))
    return (new realptr(symbtable.toid("Vel"), getattributes(), this, &v))->getlocalsymbol(name->next);

  if (*name->elem == symbtable.toid("Accel"))
    return (new realptr(symbtable.toid("Accel"), getattributes(), this, &a))->getlocalsymbol(name->next);

  if (*name->elem == symbtable.toid("ExtForce"))
    return (new realptr(symbtable.toid("ExtForce"), getattributes(), this, &extforce))->getlocalsymbol(name->next);

  if (*name->elem == symbtable.toid("PosPin"))
    return (new floatpin(symbtable.toid("PosPin"), this, getattributes(), this, &x))->getlocalsymbol(name->next);

  if (*name->elem == symbtable.toid("VelPin"))
    return (new floatpin(symbtable.toid("VelPin"), this, getattributes(), this, &v))->getlocalsymbol(name->next);

  if (*name->elem == symbtable.toid("AccelPin"))
    return (new floatpin(symbtable.toid("AccelPin"), this, getattributes(), this, &a))->getlocalsymbol(name->next);

  return (symbol *) NULL;
}

list<symbol> *motor::getapparentsymbols()
{
  int name;
  list<int> l(&name);

  name = symbtable.toid("Pos");
  getlocalsymbol(&l);

  name = symbtable.toid("Vel");
  getlocalsymbol(&l);

  name = symbtable.toid("Accel");
  getlocalsymbol(&l);

  name = symbtable.toid("ExtForce");
  getlocalsymbol(&l);

  name = symbtable.toid("PosPin");
  getlocalsymbol(&l);

  name = symbtable.toid("VelPin");
  getlocalsymbol(&l);

  name = symbtable.toid("AccelPin");
  getlocalsymbol(&l);

  return children;
}

void motor::parseattribs(attribute *attribs)
{
  attribs->getrealattribute("A", &A);
  attribs->getrealattribute("B", &B);
  attribs->getrealattribute("C", &C);
  attribs->getrealattribute("timestep", &timestep);
  if (timestep <= 0) 
  {
    error(TYPE_ERROR, -1, NULL, "Requested negative or zero timestep for motor %s, set to .0001", getfullnametext());
    timestep = .0001;
  }
}


//---------------------------------------------------------------------------

