/*
  DUST: Simulator for circuits containing a SX microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef __BREAK_H__
#define __BREAK_H__

class breakpt;

typedef int (*break_callback)(breakpt *);

class breakpt
{
  break_callback callback;
  static list<breakpt> *breaks;
  char *text; // Belongs to the breakpoint
protected:
  virtual char *_tostring() = 0;
public:
  virtual int _eval() = 0;
  static breakpt* eval();
  static void remove(breakpt *b);
  static void add(breakpt *b);
  char *tostring() { if (text == NULL) text = _tostring(); return text; }
  virtual ~breakpt() { delete text; }
  void setcallback(break_callback f) { callback = f; }
  breakpt() { callback = (break_callback) NULL; text = (char *) NULL; param = NULL; }
  void *param;
};

class breakfunc : public breakpt
{
  break_callback func;
  char *txt;
protected:
  char *_tostring() { return txt; }
public:
  int _eval() { return func(this); }
  breakfunc(break_callback f, char *text, void *param) { func = f; char *ttxt = text != NULL ? text : "@function"; text = new char[strlen(ttxt) + 1]; strcpy(text, ttxt); breakfunc::param = param; }
};

class breakop2 : public breakpt
{
  virtual char *getsymbol() = 0;
protected:
  breakpt *b1, *b2;
  char *_tostring();
public:
  ~breakop2() { delete b1; delete b2; };
  breakop2(breakpt *b1, breakpt *b2) { breakop2::b1 = b1; breakop2::b2 = b2; }
};

class breakor : public breakop2
{
  char *getsymbol() { return "||"; }
public:
  int _eval() { return b1->_eval() || b2->_eval(); } 
  breakor(breakpt *b1, breakpt *b2) : breakop2(b1, b2) {}
};

class breakand : public breakop2
{
  char *getsymbol() { return "&&"; }
public:
  int _eval() { return b1->_eval() && b2->_eval(); } 
  breakand(breakpt *b1, breakpt *b2) : breakop2(b1, b2) {}
};

class breakxor : public breakop2
{
  char *getsymbol() { return "^"; }
public:
  int _eval() { return b1->_eval() ^ b2->_eval(); } 
  breakxor(breakpt *b1, breakpt *b2) : breakop2(b1, b2) {}
};

class breakop1 : public breakpt
{
  virtual char *getsymbol() = 0;
protected:
  breakpt *b;
  char *_tostring();
public:
  ~breakop1() { delete b; };
  breakop1(breakpt *b) { breakop1::b = b; }
};

class breaknot : public breakop1
{
  char *getsymbol() { return "!"; }
public:
  int _eval() { return !b->_eval(); } 
  breaknot(breakpt *b) : breakop1(b) {}
};

class breakup : public breakop1
{
  int oldval;
  char *getsymbol() { return "+"; }
public:
  int _eval() { int oval = oldval; oldval = b->_eval(); return !oval && oldval; } 
  breakup(breakpt *b) : breakop1(b) { oldval = b->_eval(); }
};

class breakdown : public breakop1
{
  int oldval;
  char *getsymbol() { return "-"; }
public:
  int _eval() { int oval = oldval; oldval = b->_eval(); return oval && !oldval; } 
  breakdown(breakpt *b) : breakop1(b) { oldval = b->_eval(); }
};

class breakchanged : public breakop1
{
  int oldval;
  char *getsymbol() { return "~"; }
public:
  int _eval() { int oval = oldval; oldval = b->_eval(); return oval != oldval; } 
  breakchanged(breakpt *b) : breakop1(b) { oldval = b->_eval(); }
};

#define OP_EQUAL (1)
#define OP_GREATER (2)
#define OP_LOWER (4)
#define OP_GE (OP_GREATER | OP_EQUAL)
#define OP_LE (OP_LOWER | OP_EQUAL)
#define OP_DIFF (OP_GREATER | OP_LOWER)

inline char *compop(int op)
{
  switch (op)
  {
    case OP_EQUAL:
      return "==";
    case OP_GREATER:
      return ">";
    case OP_LOWER:
      return "<";
    case OP_GE:
      return ">=";
    case OP_LE:
      return "<=";
    case OP_DIFF:
      return "!=";
    default:
      return "??";
  }
};

class breakvar : public breakpt // For unsigned long!!
{
  variable *v;
  int op;
  long value;
protected:
  char *_tostring();
public:
  int _eval();
  breakvar(variable *v, long value, int op) { breakvar::v = v; breakvar::op = op; breakvar::value = value; }
};

class breakreal : public breakpt
{
  realsymbol *r;
  int op;
  real value;
protected:
  char *_tostring();
public:
  int _eval();
  breakreal(realsymbol *r, real value, int op) { breakreal::r = r; breakreal::op = op; breakreal::value = value; }
};

#endif
