/*
  DUST: Simulator for circuits containing a SX microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "project.h"

int _nullpin::nullpincount = 0;

symbolwidget *pin::getwidget()
{
  return new pinwidget(getattributes(), this);
}

pin::pin(int name, component *comp, attribute *attribs, symbol *parent) : symbol(name, attribs, parent)
{ 
  //  DEBUG(fprintf(stderr, "pin::pin(%s, , )\n", symbtable.tostring(getname()));)

  modified = 1;
  pin::comp = comp; 
  Vin = 0;
  hasparent = 0; 
}

int pin::GetModified() 
{ 
  if (!hasparent) SetVin(GetVout()); 
  return modified; 
}

real pin::GetVin()
{ 
  if (!hasparent) Vin = GetVout(); 
  //  DEBUG(cerr << "GetVin " << symbtable.tostring(getname()) << " : " << Vin << "\n";)
  return Vin; 
}

void pin::SetVin(real Vin) 
{ 
  //  DEBUG(cerr << "SetVin " << symbtable.tostring(getname()) << " : " << Vin << "\n";)

  if (Vin != pin::Vin) 
    modified = 1; 

  if ((Vin > highcarelevel && highcarelevel != +INFINITY) || (Vin < lowcarelevel && lowcarelevel != -INFINITY))
  {
    if (Project.DispOptDebugInfo)
      error(TYPE_WARNING, -1, NULL, "Pin %s modified on component %s. Vin : %.10g highcarelevel : %.10g lowcarelevel : %.10g", symbtable.tostring(getname()), comp->getfullnametext(), Vin, highcarelevel, lowcarelevel);
    comp->SetUpdateNeeded();
  }
  _SetVin(Vin); pin::Vin = Vin; 
}

void net::PropagateUp()
{
  //  DEBUG(cerr << "Propagate Up : " << symbtable.getsymbol(name) << "\n";)

  real G = 0;
  real Ivirt = 0;
  highcarelevel = INFINITY;
  lowcarelevel = -INFINITY;

  //  DEBUG(fprintf(stderr, "PropagateUp on %s\n", getfullnametext()));
  
  for (pinlist *l = pins; l != NULL; l = l->next)
  {
    if (l->elem->GetZ() == INFINITY) 
      continue;

    if (l->elem->GetZ() == 0)
    {
      SetZ(0);
      SetVout(l->elem->GetVout());

      for (pinlist *l2 = l->next; l2 != NULL; l2 = l2->next)
	if (l2->elem->GetZ() == 0 && l2->elem->GetVout() != l->elem->GetVout())
	{
	  error(TYPE_WARNING, -1, NULL, "Pins %s and %s on net %s have Z=0 and different voltages.", symbtable.tostring(l->elem->getname()), symbtable.tostring(l2->elem->getname()), getfullnametext());
	  SetVout(NAN);
	  return;
	}
      
      return;
    }

    Ivirt += l->elem->GetVout() / l->elem->GetZ();
    G += 1 / l->elem->GetZ();
    highcarelevel = MIN(highcarelevel, l->elem->highcarelevel);
    lowcarelevel = MAX(lowcarelevel, l->elem->lowcarelevel);
  }

  SetZ(1 / G);
  SetVout(Ivirt * GetZ());
}

void net::PropagateDown(int force)
{
  //  cerr << "Propagate Down : " << getfullnametext() << "\n";

  //  if (!GetModified() && !force)
 
  //    return;

    for (pinlist *l = pins; l != NULL; l = l->next)
  {
    l->elem->SetVin(GetVin());
    //    cerr << "  " << l->elem->getfullnametext() << "\n";
  }

  ClearModified();
}

net::net(int name, attribute *attribs, pinlist *pins, symbol *parent, macrocomponent *comp) : simplepin(name, comp, attribs, parent)
{ 
  net::pins = pins; 
  for (; pins != NULL; pins = pins->next)
  {
    if (pins->elem->hasparent)
      error(TYPE_ERROR, -1, NULL, "Pin %s belongs to two nets. Simulation will malfunction!", pins->elem->getfullnametext());
    pins->elem->hasparent = 1;
  }
  comp->nets = new netlist(this, comp->nets);
}

netconstructor::netconstructor(int name, attribute *attribs, list<symbolname> *pinnames)
: symbolconstructor(name, attribs) 
{ 
  DEBUG(fprintf(stderr, "netconstructor(%s, , )\n", symbtable.tostring(name));)

  netconstructor::pinnames = pinnames;
}

symbol *netconstructor::_ccreate(macrocomponent *parent)
{
  // #####Err
  pinlist *pins = (list<pin> *) NULL;

  for (list<symbolname> *l = pinnames; l != NULL; l = l->next)
  {
    symbol *s = parent->getlocalsymbol(l->elem);
    if (s == NULL) 
    {
      createerror(TYPE_ERROR, "Unknown symbol %s while creating %s. Skipped.", symbtable.tostring(l->elem), symbtable.tostring(parent->type->name));
      continue;
    }
    pin *p = s->ToPin();
    if (p == NULL) 
    {
      createerror(TYPE_ERROR, "Not a pin %s while creating %s. Skipped.", symbtable.tostring(l->elem), symbtable.tostring(parent->type->name));    
      continue;
    }
    pins = new pinlist(p, pins);
  }
  
  net *n = new net(name, attribs, pins, parent, parent);
  return n;
}

_nullpin::_nullpin(int name, attribute *attribs, component *parent) : pin(name, parent, attribs, parent)
{
   highcarelevel = +INFINITY; lowcarelevel = -INFINITY;
}

symbolwidget *_nullpin::getwidget() 
{
  return getwidget();
}

_nullpin *nullpin(attribute *attribs, component *parent)
{
  char s[] = "__nullpin__xxxxxxxxxx";
  sprintf(s, "__nullpin%i__", _nullpin::nullpincount);
  _nullpin::nullpincount++;
  return new _nullpin(symbtable.toid(s), attribs, parent);
}

outpin::outpin(int name, component *comp, attribute *attribs, symbol *parent, int initstate) : pin(name, comp, attribs, parent) 
{ 
  hilvl = 5; 
  lolvl = 0; 
  Z = 0;
  parseattribs(attribs);
  SetState(initstate); 
  highcarelevel = +INFINITY; 
  lowcarelevel = -INFINITY; 
}

list<symbol> *outpin::getapparentsymbols()
{
  int i;
  list<int> n(&i);
  i=symbtable.toid("STATE");getsymbol(&n);
  return children;
}

symbol *outpin::_getlocalsymbol(symbolname *name)
{
  if (*name->elem == symbtable.toid("STATE"))
    return (new basevariableextract<int>(symbtable.toid("STATE"), getattributes(), this, &state, 0, 0))->getlocalsymbol(name->next);

  return (symbol *) NULL;
}

void outpin::parseattribs(attribute *attribs)
{
  attribs->getrealattribute(symbtable.toid("Vcc"), &hilvl);
  attribs->getrealattribute(symbtable.toid("Vdd"), &lolvl);
  attribs->getrealattribute(symbtable.toid("Zout"), &Z);
  SetState(state);
}

list<symbol> *incomppin::getapparentsymbols()
{
  int i;
  list<int> n(&i);
  i=symbtable.toid("STATE");getsymbol(&n);
  return children;
}

symbol *incomppin::_getlocalsymbol(symbolname *name)
{
  if (*name->elem == symbtable.toid("STATE")) 
    return (new basevariableextract<int>(symbtable.toid("STATE"), getattributes(), this, &State, 0, 0))->getlocalsymbol(name->next);

  return (symbol *) NULL;
}

incomppin::incomppin(int name, component *comp, attribute *attribs, symbol *parent) : pin(name, comp, attribs, parent) 
{ 
  care = 1;
  parseattribs(attribs);
}

void incomppin::parseattribs(attribute *attribs)
{
  real hilvl = 5;
  real lolvl = 0;
  attribs->getrealattribute(symbtable.toid("Vcc"), &hilvl);
  attribs->getrealattribute(symbtable.toid("Vdd"), &lolvl);
  limit = (hilvl + lolvl) / 2;
  attribs->getrealattribute(symbtable.toid("Vtrig"), &limit);
}

void incomppin::_SetVin(real Vin)
{
  if (Vin > limit)
  {
    State = 1;
    highcarelevel = +INFINITY;
    lowcarelevel = limit;
  }
  else
  {
    State = 0;
    highcarelevel = limit;
    lowcarelevel = -INFINITY;
  }

  if (!care) 
  {
    highcarelevel = +INFINITY;
    lowcarelevel = -INFINITY;
  }
}

/****************************************************************************
Classe pin_3state
****************************************************************************/

void pin_3state::_SetVin(real Vin)
{
  if (Vin>HighLevel) {
    SetBit(Pin);
    highcarelevel=+INFINITY;
    lowcarelevel=LowLevel;
  }
  if (Vin<LowLevel) {
    ClrBit(Pin);
    highcarelevel=HighLevel;
    lowcarelevel=-INFINITY;
  }
}

real pin_3state::GetVout()
{
  if (GetBit(Tris))
    return V_OH;    //Permet de grer les pull-Ups
  else {
    if (GetBit(Latch))
      return V_OH;
    else
      return V_OL;
  }
}

real pin_3state::GetZ()
{
  if (GetBit(Tris)) {
    if GetBit(PullUp) return Z_HI;
    else return Z_PullUp;
  }
  else
    return Z_LO;
}

void pin_3state::SetTTL(int TTL)
{
  if (TTL) {  //1 =TTL
    LowLevel=V_IL_TTL;
    HighLevel=V_IH_TTL;
  } else
  {          //0=CMOS
    LowLevel=V_IL_CMOS;
    HighLevel=V_IH_CMOS;
  }
  if (GetBit(Pin)) {
    highcarelevel=+INFINITY;
    lowcarelevel=LowLevel;
  }
  else {
    highcarelevel=HighLevel;
    lowcarelevel=-INFINITY;
  }
}

list<symbol> *pin_3state::getapparentsymbols()
{
  int i;
  list<int> n(&i);
  i=symbtable.toid("STATE");getsymbol(&n);
  i=symbtable.toid("TRIS");getsymbol(&n);
  i=symbtable.toid("LATCH");getsymbol(&n);
  i=symbtable.toid("PULLUP");getsymbol(&n);
  return children;
}

symbol* pin_3state::_getlocalsymbol(symbolname *name)
{
  int id;

  if (*(name->elem) == (id=symbtable.toid("STATE")) ) {
    int b=log2(Pin.Mask);
    return (new basevariableextract<char>(id, getattributes(), this, Pin.Address, b, b ))->getlocalsymbol(name->next);
  }
  if (*(name->elem) == (id=symbtable.toid("TRIS")) ) {
    int b=log2(Tris.Mask);
    return (new basevariableextract<char>(id, getattributes(), this, Tris.Address, b, b ))->getlocalsymbol(name->next);
  }
  if (*(name->elem) == (id=symbtable.toid("LATCH")) ) {
    int b=log2(Latch.Mask);
    return (new basevariableextract<char>(id, getattributes(), this, Latch.Address, b, b ))->getlocalsymbol(name->next);
  }
  if (*(name->elem) == (id=symbtable.toid("PULLUP")) ) {
    int b=log2(PullUp.Mask);
    return (new basevariableextract<char>(id, getattributes(), this, PullUp.Address, b, b ))->getlocalsymbol(name->next);
  }
  return (symbol*)NULL;
}



/****************************************************************************
Classe pin_output
****************************************************************************/

pin_Output::pin_Output(int name, component *comp, attribute *attribs, symbol *parent
  ,char* Address,char Mask):pin(name,comp,attribs,parent)
{
  Latch.Address=Address;
  Latch.Mask=Mask;
  highcarelevel=INFINITY;  // broche en sortie, ne ncessite pas d'update
  lowcarelevel=-INFINITY;
}

real pin_Output::GetVout()
{
  if (GetBit(Latch))
    return V_OH;
  else
    return V_OL;
}

real pin_Output::GetZ()
{
  return Z_LO;
}

list<symbol> *pin_Output::getapparentsymbols()
{
  int i;
  list<int> n(&i);
  i=symbtable.toid("LATCH");getsymbol(&n);
  return children;
}

symbol* pin_Output::_getlocalsymbol(symbolname *name)
{
  int id;

  if ((*(name->elem) == (id=symbtable.toid("LATCH")) )||*(name->elem) == (symbtable.toid("STATE"))) {
    int b=log2(Latch.Mask);
    return (new basevariableextract<char>(id, getattributes(), this, Latch.Address, b, b ))->getlocalsymbol(name->next);
  }
  return (symbol*)NULL;
}

/****************************************************************************
Classe pin_Edge
****************************************************************************/

pin_Edge::pin_Edge(int name, component *comp, attribute *attribs, symbol *parent
  ,char* Address,char Mask):pin_Input(name,comp,attribs,parent,Address,Mask)
{
  Detect=0;
}

void pin_Edge::_SetVin(real Vin)
{
  char LastValue=GetBit(Pin);
  pin_Input::_SetVin(Vin);
  if (Edge) {
    if ((LastValue)&&(!GetBit(Pin))) Detect=1;
  } else {
    if ((!LastValue)&&(GetBit(Pin))) Detect=1;
  }
}

list<symbol> *pin_Edge::getapparentsymbols()
{
  pin_Input::getapparentsymbols();
  int i;
  list<int> n(&i);
  i=symbtable.toid("DETECT");getsymbol(&n);
  i=symbtable.toid("EDGE");getsymbol(&n);
  return children;
}

symbol* pin_Edge::_getlocalsymbol(symbolname *name)
{
  int id;
  symbol* Temp=pin_Input::_getlocalsymbol(name);
  if (!Temp) {
    if (*(name->elem) == (id=symbtable.toid("DETECT")) ) {
      return (new basevariableextract<char>(id, getattributes(), this, &Detect, 0, 0 ))->getlocalsymbol(name->next);
    }
    if (*(name->elem) == (id=symbtable.toid("EDGE")) ) {
      variable* var=(new basevariableextract<char>(id, getattributes(), this, &Edge, 0, 0 ));
      var->readonly=1;
      return var->getlocalsymbol(name->next);
    }
  }
  return Temp;
}

/****************************************************************************
Classe pin_input
****************************************************************************/
pin_Input::pin_Input(int name, component *comp, attribute *attribs, symbol *parent
  ,char* Address,char Mask):pin(name,comp,attribs,parent)
{
  Pin.Address=Address;
  Pin.Mask=Mask;
  HighLevel=V_IH_I;
  LowLevel=V_IL_I;
}

void pin_Input::_SetVin(real Vin)
{
  if (Vin>HighLevel) {
    SetBit(Pin);
    highcarelevel=+INFINITY;
    lowcarelevel=LowLevel;
  }
  if (Vin<LowLevel) {
    ClrBit(Pin);
    highcarelevel=HighLevel;
    lowcarelevel=-INFINITY;
  }
}

real pin_Input::GetVout()
{
  return 0;
}

real pin_Input::GetZ()
{
  return Z_HI;
}

list<symbol> *pin_Input::getapparentsymbols()
{
  int i;
  list<int> n(&i);
  i=symbtable.toid("STATE");getsymbol(&n);
  return children;
}

symbol* pin_Input::_getlocalsymbol(symbolname *name)
{
  int id;

  if (*(name->elem) == (id=symbtable.toid("STATE")) ) {
    int b=log2(Pin.Mask);
    return (new basevariableextract<char>(id, getattributes(), this, Pin.Address, b, b ))->getlocalsymbol(name->next);
  }
  return (symbol*)NULL;
}

//---------------------------------------------------------------------------
