/*
  DUST: Simulator for circuits containing a SX microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef __VARIABLE_H__
#define __VARIABLE_H__

// variables are limited to long for convienience. I don't see a practical use for more
// that justifies using long integers.

char *getstringvalue(unsigned int value, int size, int base);
char *getsignedstringvalue(int value, int size, int base);
int setstringvalue(char *strvalue, int defbase, unsigned long* result);

class variable : public symbol
{
public:
  bool readonly;
  virtual symbolwidget *getwidget();

  variable *ToVariable() { return this; }

  virtual unsigned long getvalue() = 0;
  long getsignedvalue() { int remains = sizeof(long) * 8 - getsize(); return ((long) getvalue() << remains) >> remains; }
  char *getstringvalue(int defbase);
  char *getsignedstringvalue(int defbase);
  virtual void _setvalue(unsigned long) {};
  void setvalue(unsigned long i) { if (!readonly) _setvalue(i); };
  int setvalue(char *strval, int base);
  virtual int getsize() = 0;
  virtual unsigned long getbits(int start, int end) = 0; // start and end are included
  virtual void setbits(int start, int end, unsigned long value) = 0;

  variable(int name, attribute *attribs, symbol *parent) : symbol(name, attribs, parent) { readonly = 0; }
};

// Applies to builtin types. Necessary to avoid exceptions if we want to access
// a byte on the edge of allowed memory zone. Best use unsigned.

template <class t> class basevariable : public variable
{
  t *ptr;
public:
  virtual unsigned long getvalue() { return *ptr; }
  virtual void _setvalue(unsigned long val) { *ptr = val; };
  virtual int getsize() { return sizeof(t) * 8; }
  virtual unsigned long getbits(int start, int end) { return ((*ptr) & ~((~1) << end)) >> start; }
  virtual void setbits(int start, int end, unsigned long value);

  basevariable(int name, attribute *attribs, symbol *parent, t *ptr) : variable(name, attribs, parent) { basevariable::ptr = ptr; }
};

// If you are stupid enough to ask for out of range values and expect some good to come
// from it then you deserve to get what comes. The parsing process ensures that these
// values vill be valid.

template <class t> class basevariableextract : public basevariable<t>
{
  int _start; // First included bit
  int _end; // Last included bit
public:
  virtual unsigned long getvalue() { return basevariable<t>::getbits(_start, _end); }
  virtual void _setvalue(unsigned long val) { basevariable<t>::setbits(_start, _end, val); };
  virtual int getsize() { return 1 + _end - _start; }
  virtual unsigned long getbits(int start, int end) { return basevariable<t>::getbits(_start + start, _start + end);  }
  virtual void setbits(int start, int end, unsigned long val) { basevariable<t>::setbits(_start + start, _start + end, val);  }

  basevariableextract(int name, attribute *attribs, symbol *parent, t *ptr, int start, int end) : basevariable<t>(name, attribs, parent, ptr) { _start = start; _end = end; }
};

struct fusionvariableelement
{
  variable *var;
  int start;
  int end;
};

class fusionvariable : public variable
{
  int bits;
  list<fusionvariableelement> *vars; // Most significant first
public:
  virtual unsigned long getvalue();
  virtual void _setvalue(unsigned long val) { setbits(0, bits - 1, val); };
  virtual int getsize() { return bits; }
  virtual unsigned long getbits(int start, int end) { return ((getvalue()) & ~((~1) << end)) >> start;}
  virtual void setbits(int start, int end, unsigned long value);

  fusionvariable(int name, attribute *attribs, symbol *parent, list<fusionvariableelement> *vars);
};

// Only applies to fusion variables

class variableconstructor : public symbolconstructor
{
  list<symbolname> *pieces;

protected:
  virtual symbol *_create(symbol *parent);

public:
  variableconstructor(int name, attribute *attribs, list<symbolname> *pieces) : symbolconstructor(name, attribs) { variableconstructor::pieces = pieces; }
};

#endif
